getBlockList <- function(locs,n1,n2){
    
    # blocks the rectangle enclosing locs into a 
    # n1 x n2 grid, and returns a list of the indices
    # of locs that fall in each block
    
    r1 <- range(locs[,1])
    r2 <- range(locs[,2])
    locsround <- matrix(0,n,2)
    locsround[,1] <- floor( (locs[,1] - r1[1])/diff(r1)*(n1-1e-6) + 1)
    locsround[,2] <- floor( (locs[,2] - r2[1])/diff(r2)*(n2-1e-6) + 1)
    blocklist <- vector("list",length=n1*n2)
    for(j1 in 1:n1){
        for(j2 in 1:n2){
            blocklist[[(j1-1)*n2 + j2]] <- which(locsround[,1]==j1 & locsround[,2]==j2)     
        }
    }
    blocklist
}

mvnMargLikInds <- function(inds,covparms,covfun,y,locs){
    exactLik(covparms,covfun,y[inds],locs[inds,])
}

mvnIndepBlocks <- function(covparms,covfun,y,locs,blocklist){
    
    # blocklist is a list. each list element is a vector 
    # of indices, defining the blocks.
    liklist <- lapply(blocklist,mvnMargLikInds,
                      covparms=covparms,covfun=covfun,y=y,locs=locs)
    return(sum(unlist(liklist)))
}

findOrderedNNdistant <- function( locs, m, pnear = 1 ){
    # find the pnear(m+1) nearest neighbors to locs[j,] in locs[1:j,]
    # by convention, this includes locs[j,], which is distance 0
    
    # also adds (1-pnear)(m+1) points from the distant past
    # for a total of m+1 neighbors
    
    n <- dim(locs)[1]
    NNarray <- matrix(NA,n,m+1)
    mnear <- floor(pnear*(m+1))
    mfar <- m+1 - mnear
    
    for(j in 1:n ){
        distvec <- c(rdist(locs[1:j,,drop=FALSE],locs[j,,drop=FALSE]) )
        odist <- order(distvec)
        if( j <= m+1) NNarray[j,1:j] <- odist[1:j]        
        if( j > m+1){
            NNarray[j,1:mnear] <- odist[1:mnear]
            if( mfar > 0 ){
                # have to double check to make sure this has no duplicates
                # (and that it's doing what I think it is
                leftinds <- (mnear+1):j
                inds <- floor( seq(mnear+1,j,length.out=mfar) ) 
                NNarray[j,(mnear+1):(m+1)] <- odist[inds]
            }
        }
    }    
    NNarray
}



spdefun <- function( locs,range0,tau0,cutoff=1e-6,offset=c(0.2,0.2),max.edge=c(0.2,0.2), getKL=FALSE, covmat=NULL, cholmat=NULL ){
    
    # construct spde mesh
    mesh = inla.mesh.2d(
        loc=locs,
        cutoff = cutoff,
        offset=offset,
        max.edge=max.edge)
    
    n <- nrow(locs)
    
    kappa0 = sqrt(8)/range0
    
    ## Construct spde object
    spde=inla.spde2.matern(mesh, alpha=2,
                           B.tau=cbind(log(tau0),1,0),
                           B.kappa=cbind(log(kappa0),0,1),
                           theta.prior.mean=c(0,0),
                           theta.prior.prec=c(0.1,1))
    
    # get precision matrix Q 
    Q=inla.spde2.precision(
        spde,theta=c(0,0))
    
    # compute KLdivergence if indicated
    if(getKL){
        allind <- 1:dim(Q)[1]
        idx1 <- mesh$idx$loc   # observation indices
        idx2 <- allind[-idx1]
        Q11 <- Q[idx1,idx1]
        Q12 <- Q[idx1,idx2]
        Q21 <- Q[idx2,idx1]
        Q22 <- Q[idx2,idx2]
        siginv <- Q11 - Q12 %*% solve(Q22,Q21)
        #kldiv <- -sum(log(diag(cholmat)))-sum(log(diag(chol(siginv)))) + 1/2*sum(diag(siginv %*% covmat)) - n/2 
        kldiv <- -sum(log(diag(cholmat)))-sum(log(diag(chol(siginv)))) + 1/2*( sum( siginv * covmat ) - n ) 
        return(kldiv)
    } else { # otherwise just return the precision
        return(Q)
    }
}                 



taperfun <- function( locs, varparm, rangeparm, taperrange, taperk, covmat ){
    tapercov <- varparm*stationary.taper.cov(locs,theta=rangeparm,
                                             Taper.args = list(theta=taperrange,k=taperk)) 
    choltaper <- chol.spam(tapercov)
    invtapercov <- solve(as.matrix(tapercov))
    #    mat <- backsolve( choltaper, covmat )
    #    mat <- forwardsolve( t(choltaper), mat )
    #    kldiv <- 1/2*(sum(diag(solve( tapercov, covmat ))) - n) + 
    #        sum(log(diag(choltaper))) + (logdet0 + n/2*log(2*pi))
    #    kldiv <- 1/2*(sum(diag(mat)) - n) + 
    #        sum(log(diag(choltaper))) + (logdet0 + n/2*log(2*pi))
    kldiv <- 1/2*(sum(invtapercov * covmat) - n) + 
        sum(log(diag(choltaper))) + (logdet0 + n/2*log(2*pi))
    
    return(kldiv)
}


exactLik <- function(covparms,covfun,y,locs){
    
    n <- length(y)
    covmat <- covfun(locs,covparms)
    cholmat <- t(chol(covmat))
    z <- forwardsolve(cholmat,y)
    quadform <- sum( z^2 )
    logdet <- 2*sum(log(diag(cholmat)))
    loglik <- -1/2*( n*log(2*pi) + logdet + quadform )
    return(loglik)
    
}


groupNN <- function(NNarray){
    n <- nrow(NNarray)
    m <- ncol(NNarray)-1

    clust <- vector("list",n)
    for(j in 1:n) clust[[j]] <- j
    for( ell in 2:(m+1) ){  # 2:(m+1)?
        sv <- which( NNarray[,1] - NNarray[,ell] < n )
        for(j in sv){
            k <- NNarray[j,ell]
            if( length(clust[[k]]) > 0){
                nunique <- length(unique(c(NNarray[c(clust[[j]],clust[[k]]),])))

                # this is the rule for deciding whether two groups
                # should be combined
                if( nunique^2 <= length(unique(c(NNarray[clust[[j]],])))^2 + length(unique(c(NNarray[clust[[k]],])))^2 ) {
                    clust[[j]] <- c(clust[[j]],clust[[k]])
                    clust[[k]] <- numeric(0)
                }
            }
        }
    }
    zeroinds <- unlist(lapply(clust,length)==0)
    clust[zeroinds] <- NULL
    NNlist <- lapply(clust,function(inds) NNarray[inds,,drop=FALSE])
    return(NNlist)
}


simulateGrid <- function(nvec,errorsize=0){
    n <- prod(nvec)
    d <- length(nvec)
    gridlocs <- createGrid(nvec)
    u <- matrix(runif(n*d)-1/2,n,d) %*% diag(errorsize/nvec)
    locs <- gridlocs + u
}


createGrid <- function(nvec){
    
    if(missing(nvec) || length(nvec) == 0) stop("Must supply grid dimensions")
    d <- length(nvec)
    n <- prod(nvec)
    gridlocs <- matrix(0,n,d)
    for(j in 1:d){
        perm <- (1:d)[-j]
        perm <- c(j,perm)
        tempvec <- nvec[-j]
        tempvec <- c(nvec[j],tempvec)
        a1 <- ((1:nvec[j])-1/2)/nvec[j]
        a2 <- rep(a1,n/nvec[j])
        a3 <- array(a2,tempvec)
        a4 <- aperm(a3,perm)
        gridlocs[,j] <- c(a4)
    }
    gridlocs
}




clusterNN2 <- function(NNarray){
    n <- nrow(NNarray)
    m <- ncol(NNarray)-1
    
    clust <- vector("list",n)
    for(j in 1:n) clust[[j]] <- j
    for( ell in 2:(m+1) ){  # 2:(m+1)?
        sv <- which( NNarray[,1] - NNarray[,ell] < n )
        for(j in sv){
            k <- NNarray[j,ell]
            if( length(clust[[k]]) > 0){
                nunique <- length(unique(c(NNarray[c(clust[[j]],clust[[k]]),])))
                if( nunique^2 <= length(unique(c(NNarray[clust[[j]],])))^2 + length(unique(c(NNarray[clust[[k]],])))^2 ) {
                    clust[[j]] <- c(clust[[j]],clust[[k]])
                    clust[[k]] <- numeric(0)
                }
            }
        }
    }
    zeroinds <- unlist(lapply(clust,length)==0)
    clust[zeroinds] <- NULL
    NNlist <- lapply(clust,function(inds) NNarray[inds,,drop=FALSE])
    return(NNlist)
}


